#!/usr/bin/env python3

import xml.etree.ElementTree as ET
from nltk.tokenize import RegexpTokenizer
from nltk.corpus import stopwords
from nltk.metrics import edit_distance
from pyxdameraulevenshtein import damerau_levenshtein_distance_ndarray, normalized_damerau_levenshtein_distance_ndarray
import numpy as np
import pandas as pd
import string
import re
import os, sys
sys.path.append('src/python/')
import utils_proc
import copy
import collections


# function to check whether a file containts discussions
# achieved by excluding title pages, table of content, etc.
# !!! function works well for 1891 - 1900, not checked after that !!!
def check_if_discussion(path_meta_xml_file,
        list_attributes  = ['TITEL_NORMAL_DE', 'TITEL_NORMAL_FR', 'TITEL_ORIGINAL_DE', 'TITEL_ORIGINAL_FR'],
        list_nondiscussion = ['inhaltsverzeichnis', 'inhaltsverzeichniss', 'jahresinhalt', 'rednerliste',
            'umschlag', 'sachregister', 'titelblatt', 'numerierung'],
        list_nondiscussion2 = ['table', 'matières', 'répertoire', 'procès-verbaux']):

    # parse, get root and then part of interest
    XML_tree = ET.parse(path_meta_xml_file)
    XML_root = XML_tree.getroot()
    XML_poi = XML_root[0].find('ADS_TEXTEINHEIT')

    # for each title attribute
    for attribute in list_attributes:
        # if xml contains this attribute
        if attribute in XML_poi.attrib:
            # get title and generate set with lower case terms
            title = XML_poi.attrib[attribute]
            set_title = set([term.lower() for term in title.split()])
            #print(set_title)
            # if one of terms is in list_nondiscussion, return False
            if set_title.intersection(set(list_nondiscussion)):
                #print('NOOO', path_meta_xml_file)
                return False
            # if two terms are in list_nondiscussion2, also return False
            if len(set_title.intersection(set(list_nondiscussion2))) > 1:
                #print('NOOO', path_meta_xml_file)
                return False

    return True

# helper function to get text without font information
# example for font information: [font face="11.718" size="Times-Roman"] sometext [/font]
# input:
# - sometext: string
# output:
# - newtext: modified string
def get_text(sometext):
    # initialize
    newtext = ''

    # find text between font information
    for text in re.findall('\].*?\[',sometext):
        #print(text)
        if text.startswith(']') and text.endswith('['):
            newtext += text[1:-1]
    #print(newtext)
    return newtext


# function to annotated corrected XML
def get_annotated_xml(XML_root, df_lastnames, bln_print=False):

    # list of votation terms
    # TODO: make it work for é, etc.
    list_votationterms = ['Abstimmung', 'Schlussabstimmung', 'Generalabstimmung', 'Angenommen', 'Abgelehnt',
                          'Einverstanden', 'Stimmen', '(Eintreten)', '(Nichteintreten)',
                          'Votation', 'Vote', 'votation', #'(Adopt�s)', 'adopt�s', 'adopt�e', 'rejet�e',
                          "D'accord", 'voix']

    # list of stopwords
    list_stopwords = stopwords.words('german')
    list_stopwords.extend(['dass', 'resp', 'daran', 'dr', 'herr', 'herrn', 'hr'])
    list_stopwords.extend(stopwords.words('french'))
    list_stopwords.extend(['ils', 'les', 'celle'])

    # create new XML as a copy of the corrected one
    XML_new = copy.deepcopy(XML_root)

    # initialize flags to distinguish speeches from votes
    this_is_speech = False
    prev_is_speech = False
    this_is_vote = False

    # for every page
    for ind_p, page in enumerate(XML_root):
        if bln_print:
            print(page.tag, page.attrib)
        # for every textbox on that page
        for ind_t, textbox in enumerate(page):
            if (textbox.tag == 'textbox'):
                if 'type_textbox' in textbox.attrib.keys():
                    if (textbox.attrib['type_textbox'] == 'text'):
                        if bln_print:
                            print(textbox.tag, textbox.attrib)

                        # get complete text of that textbox
                        complete_text, ind_tl_colon = get_complete_text(textbox)
                        if bln_print:
                            print(complete_text[:100])

                        # identify and label language in XML
                        dict_lang = identify_language(complete_text)
                        XML_new = label_language(XML_new, ind_p, ind_t, dict_lang)

                        # get texttype of that textbox by majority vote
                        # TODO add that type to XML
                        textbox_texttype = get_textbox_type(textbox)
                        if bln_print:
                            print(textbox_texttype)

                        if textbox_texttype in ['text_col1', 'text_col2']:

                            XML_new, this_is_speech = label_speechstart(XML_new, ind_p, ind_t, complete_text, ind_tl_colon, df_lastnames, list_stopwords, bln_print=False)
                            if this_is_speech:
                                prev_is_speech = True
#                                print('stopped after finding speech start')
                                continue
                            XML_new, this_is_vote = label_votations(XML_new, ind_p, ind_t, complete_text, list_votationterms, bln_print=False)
                            if this_is_vote:
                                prev_is_speech = False
#                                print('stopped after finding vote')
                                continue
                            if prev_is_speech and (not this_is_vote):
                                XML_new = label_speechcont(XML_new, ind_p, ind_t)


    return XML_new


# helper function to get type of textbox_type
# corresponds to majority vote of types of textlines
# input:
# - textbox
# output:
# - textbox_type: string
def get_textbox_type(textbox):

    # initialize empty dictionary
    dict_type = {}

    # for every textline in that textbox
    for ind_tl, textline in enumerate(textbox):
        if textline.tag == 'textline':

            # count types
            if textline.attrib['type'] not in dict_type.keys():
                dict_type[textline.attrib['type']] = 1
            else:
                dict_type[textline.attrib['type']] += 1

    # list of all types with maximum count
    list_types = [type for type, count in dict_type.items() if count == max(dict_type.values())]

    # if only one with maximum value
    if len(list_types) == 1:
        textbox_type = list_types[0]

    # if several with same maximum value
    else:
        textbox_type = 'notdistinct'

    return textbox_type


# helper function to get complete text of a textbox
# input:
# - textbox
# output:
# - complete_text: string
# - ind_tl_colon: index of textline with colon (needed for label speech start)
def get_complete_text(textbox):

    # helper function to get text without font information
    def get_text(sometext):
        newtext = ''
        for text in re.findall('\].*?\[',sometext):
            #print(text)
            if text.startswith(']') and text.endswith('['):
                newtext += text[1:-1]
        #print(newtext)
        return newtext

    # initialize empty string
    complete_text = ''

    # initialize index of textline colon to impossible value
    ind_tl_colon = -1

    # for every textline in that textbox
    for ind_tl, textline in enumerate(textbox):
        if textline.tag == 'textline':
            # get that text
            thattext = get_text(textline.text)

            # append that text to string
            complete_text += thattext

            # in first two textlines of textbox, check for colon
            if ind_tl < 3:
                if ':' in thattext:
                    ind_tl_colon = ind_tl

    return complete_text, ind_tl_colon


# function to label speech starts
# input:
# - text: string to be analyzed
# - df_names: dataframe of politicians
# - list_stopwords: list of german and french stopwords
# - bln_print: whether to print during execution, default False
# output:
# - (str_name, str_role, list_uniqueID, str_canton): tuple with strings and ID
def label_speechstart(XML_new, ind_p, ind_t, text, ind_tl_colon, df_names, list_stopwords, bln_print=False):

    # initialize flag
    this_is_speech = False

    # initialize strings and ID
    str_name = ''
    str_role = ''
    list_uniqueID = []
    str_canton = ''

    # font text end
    fontend = '[/font]'

    # very consistently, a speaker can be identified by looking for a colon
    # at the beginning of a textbox and identifiying a name or a role in front
    # of that colon
    if ind_tl_colon >= 0:
#    if ':' in text[:100]:
        # extract the index of the colon in the text
        colon_index_text = text.index(':')

        # look at first few terms of that textbox
        text_start = re.sub(r'[\(\)]','',text[:colon_index_text])
        list_oi = tokenizer.tokenize(text_start)
        if bln_print:
            print('possible speech start: ', list_oi)

        # remove stopwords
        list_oi = [term for term in list_oi if term.lower() not in list_stopwords]

        # remove punctuation
        list_oi = [''.join(c for c in s if c not in string.punctuation) for s in list_oi]
        list_oi = [s for s in list_oi if s]

        # remove lower case terms
#        list_oi = [term for term in list_oi if not term.islower()]

        # remove numbers
        list_oi = [term for term in list_oi if not term.isdigit()]

        # remove single characters
        list_oi = [term for term in list_oi if len(term)>1]

        # for every term
        for term in list_oi:
            # if possible, find a name in a list
            str_name, str_role, list_uniqueID, str_canton = find_names(term, str_name, str_role, list_uniqueID, df_names, str_canton, bln_print=False)
        if bln_print:
            print('name', str_name, 'role', str_role)

        # get rid of doubled double names
        # TODO

        # get rid of 'Präsident stimmt nicht Président ne vote pas'
        if set(str_role.split()).intersection(set(['Präsident', 'Präsidentin', 'Président', 'Présidente'])) and not str_name:
            if set(['stimmt', 'nicht', 'vote', 'pas']).intersection(list_oi):
                if bln_print:
                    print('get rid of Präsident stimmt nicht, Président ne vote pas', list_oi)
                str_role = ''

        # get rid of 'Für den Antrag "Name" stimmen: Votent pour la proposition "Name":'
        if str_name:
            if len(set(['Antrag', 'stimmen', 'Votent', 'proposition']).intersection(list_oi)) > 1:
                if bln_print:
                    print('get rid of Für den Antrag <Name> stimmen: Votent pour la proposition <Name>:', list_oi)
                str_name = ''

        # if a name has been found, add it to XML_new
        if str_name or str_role:
            # add attribute speech_start to textbox
            XML_new[ind_p][ind_t].attrib['text_type'] = 'speech_start'

            # add speaker as attribute to first textline
            XML_new[ind_p][ind_t][0].attrib['speaker'] = (str_name, str_role, list_uniqueID, str_canton)

            # update text of XML (speaker is on first line, actual speech start on second line of speech_start textbox)
            # if colon is on first line
            if ind_tl_colon == 0:
                # get text of that line and colon index
                thattext = XML_new[ind_p][ind_t][0].text
                colon_index = thattext.index(':')

                # get last font information of thattext
                fontstart = re.findall('\[font.*?\]', thattext)[-1]

                try:
                    # write speaker to first line
                    XML_new[ind_p][ind_t][0].text = thattext[:colon_index+1] + fontend

                    # get start of speech with correct font start
                    if thattext[colon_index+1:].startswith('[font'):
                        startspeech = thattext[colon_index+1:]
                    elif re.match('^[ ]?\[/font\]', thattext[colon_index+1:]):
                        startspeech = ''
                    else:
                        startspeech = fontstart + thattext[colon_index+1:]

                    # write beginning of speech to second line
                    # (create new ET element if necessary)
                    if len(list(XML_new[ind_p][ind_t])) > 1:
                        XML_new[ind_p][ind_t][1].text = startspeech + ' ' + XML_new[ind_p][ind_t][1].text
                    else:
                        XML_new[ind_p][ind_t].append(copy.deepcopy(XML_new[ind_p][ind_t][0]))
                        XML_new[ind_p][ind_t][1].attrib.pop('speaker')
                        XML_new[ind_p][ind_t][1].text = startspeech
                except:
                    print('error in self.input_file when splitting speaker')
                    #print(thattext)
                    #print(len(list(XML_new[ind_p][ind_t])))
                    #print(list(XML_new[ind_p][ind_t]))
                    #print(XML_new[ind_p][ind_t])
                    #print('gefundener Name:', str_name, str_role)
                    pass

            # if colon is on second line
            if ind_tl_colon == 1:
                # get text of that line and colon index
                thattext = XML_new[ind_p][ind_t][1].text
                colon_index = thattext.index(':')

                # get last font information of thattext
                fontstart = re.findall('\[font.*?\]', thattext)[-1]

                # get start of speech with correct font start
                if thattext[colon_index+1:].startswith('[font'):
                    startspeech = thattext[colon_index+1:]
                elif re.match('^[ ]?\[/font\]', thattext[colon_index+1:]):
                    startspeech = ''
                else:
                    startspeech = fontstart + thattext[colon_index+1:]

                # write speaker to first line
                XML_new[ind_p][ind_t][0].text = XML_new[ind_p][ind_t][0].text + ' ' + thattext[:colon_index+1] + fontend
                # write beginning of speech to second line
                XML_new[ind_p][ind_t][1].text = startspeech

            # set flag
            this_is_speech = True
            if bln_print:
                print('found a name:', list_oi, str_name, str_role, '\n')

    return XML_new, this_is_speech


# function to extract votation paragraphs
# !!! error prone, possible improvements see notebook extract_discussions
# input:
# - XML_new:
# - text: string
# - list_votationterms: list of votation terms
# - bln_print: whether to print during execution, default False
# output:
# - XML_new: updated
def label_votations(XML_new, ind_p, ind_t, text, list_votationterms, bln_print=True):

    # get first terms of that text
    list_oi = tokenizer.tokenize(text)[:15]
#        if len(set(list_oi).intersection(set(list_votationterms))) > 1:
    # if there is an overlap with typical votation terms:
    if set(list_oi).intersection(set(list_votationterms)):
        # add attribute vote to textbox
        XML_new[ind_p][ind_t].attrib['text_type'] = 'vote'

        # set flag
        this_is_vote = True
        if bln_print:
            print('found a vote:', list_oi)
    else:
        #pass
        # set flag
        this_is_vote = False
        if bln_print:
            print('not a vote', list_oi)

    return XML_new, this_is_vote


# function to label continuation of speech
# only adds label to corresponding textbox
def label_speechcont(XML_new, ind_p, ind_t):

    XML_new[ind_p][ind_t].attrib['text_type'] = 'speech_cont'

    return XML_new

# helper function to flatten nested irregular list
def flatten(l):
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el

# function to find names
# input:
# - term: term that might be name
# - str_name: string to which name should be attached
# - str_role: string to which role should be attached
# - list_uniqueID: list with one or several uniqueIDs
# - list_tupels: list of tupels containing all types of names
# TODO: correctly extract canton! don't do reversed, find name first that might have issue with canton, then look for possible canton
# TODO: double names in extra column: string gets written twice, uniqueID is fine (1893, 20026532, Cramer)
def find_names(term, str_name, str_role, list_uniqueID, df_names, str_canton, bln_print=False):

    def get_string(term, str_name, str_role, list_uniqueID, str_canton):
        name_type = ''
        # if it is one of the simple names
        if term in list(df_names['name_short'].loc[df_names['type']=='simple']):
            str_name = add_to_string(str_name, term)
            name_type = 'simple'
        # if it is a role
        elif term in list_roles:
            str_role = add_to_string(str_role, term)
        # if it is a double name
        elif term in list(df_names['name_short'].loc[df_names['type']=='double']):
            if bln_print:
                print(20*'\n', 'DOUBLE NAME')
            # get correct name
            correct_name = df_names.loc[(df_names['type']=='double') & (df_names['name_short']== term)].iat[0, df_names.columns.get_loc('name_correct')]
            if bln_print:
                print('double name', correct_name)
            # only add name if it is not there yet
            # if a person is referenced by its complete double name, e.g. Meier-Müller, he or she gets two entries
            if correct_name not in str_name.split(' '):
                str_name = add_to_string(str_name, correct_name)
            name_type = 'double'
        # if it is a composite name
        elif term in list(df_names['name_short'].loc[df_names['type']=='comp']):
            # get correct name
            correct_name = df_names.loc[(df_names['type']=='comp') & (df_names['name_short']== term)].iat[0, df_names.columns.get_loc('name_correct')]
            if bln_print:
                print('composite name', correct_name)
            str_name = add_to_string(str_name, correct_name)
            name_type = 'comp'
        # if it contains a canton
        # !!! also pass list_oi to look for canton
        # !!! how to handle for people mentioned in text???
        elif term in list(df_names['name_short'].loc[df_names['type']=='canton']):
            if bln_print:
                print('contains a canton', term)
#            canton_missing = False
#            df_temp = df_names.loc[df_names['name_short']==term]
#            print('list_correct', df_temp)
#            print(str_canton)
#            if str_canton:
#                str_correct = check_place(df_temp, str_canton)
#                if str_correct in ['not found', 'too many']:
#                    str_name = add_to_string(str_name, term + ' (CANTONT MISSING)')
#                    canton_missing = True
#                else:
#                    str_name = add_to_string(str_name, str_temp)
#                    name_type = 'canton'
            str_name = add_to_string(str_name, term + ' (CANTON MISSING)')
            name_type = 'canton'


        # extract uniqueID
        list_temp = []
        if name_type in ['simple', 'double', 'comp']:
            list_temp = [df_names.loc[(df_names['type']==name_type) & (df_names['name_short']==term)].iat[0, df_names.columns.get_loc('uniqueIndex')]]
        elif name_type in ['canton']:
            list_temp = list(df_names.loc[(df_names['type']==name_type) & (df_names['name_short']==term)].iloc[:, df_names.columns.get_loc('uniqueIndex')])
#            if canton_missing:
#                temp = tuple(df_names.loc[(df_names['type']==name_type) & (df_names['name_short']==term)].iloc[:, df_names.columns.get_loc('uniqueIndex')])
#            else:
#                temp = df_names.loc[(df_names['type']==name_type) & (df_names['name_correct']==str_correct)].iat[0, df_names.columns.get_loc('uniqueIndex')]

        if len(list_temp) > 0:
            if bln_print:
                print(list_temp, list_uniqueID)
                print(type(list_temp), type(list_uniqueID))
                print(isinstance(list_uniqueID, list))
            # if no unique ID has been assigned so far
            if len(list_uniqueID) == 0:
                list_uniqueID = list_temp
            # if there are already one or several people and have a new person, we update
            elif len(list_uniqueID) > 0 and len(set(list_temp).intersection(set(flatten(list_uniqueID)))) == 0:
                list_uniqueID.append(list_temp)

            ## if we already have several possible people, e.g. because of canton
            #elif isinstance(int_uniqueID, tuple):
                #print('I should be here')
                ## and refound the uniqueID of one of those, don't update
                #if temp in int_uniqueID:
                    #pass
                ## and update if we don't have that uniqueID yet
                #else:
                    #int_uniqueID = (int_uniqueID, temp)
            ## if a person with that uniqueID exists already, don't update
            #elif isinstance(int(int_uniqueID), int) and temp == int_uniqueID:
                #print('but end up here.. not even.....')
                #pass
            ## if a different unique ID has been assigned already
            #else:
                #int_uniqueID = (int_uniqueID, temp)

        return str_name, str_role, list_uniqueID

    # small function to add term to str_name
    def add_to_string(string, term):
        if not string:
            string = term
        else:
            string += ' ' + term
        return string

    list_roles = ['Präsident', 'Präsidentin', 'Vizepräsident', 'Präsidium', 'Président', 'Présidente', 'président', 'présidente',
                  'Berichterstatter', 'Berichterstatterin', 'rapporteur',
                  'Sprecher', 'Sprecherin', 'porte-parole', 'porteparole',
                  'Bundesrat', 'Bundesrath', 'Bundesrätin', 'conseiller fédéral',
                  'Vizepräsident']

    list_notnames = ['Art', 'Rath', 'Alinea', 'Stimmen', 'Stimme', 'Hans', 'Walter', 'Werner', 'projet', 'stimmt', 'nicht', 'vote', 'Gallen', 'StGallen',
                     'Kasse', 'fasse', 'Sitten', 'Herren', 'Herr', 'Alter', 'Biffer', 'biffer', 'Rédiger', 'rédiger', 'Wer', 'Fällen']

    list_places = get_list_cantons(df_names)

    if bln_print:
        print('now is about: ------', term)
    # extract list and array of last names
    list_all_names = list(df_names['name_short'])
    array_all_names = np.array(df_names['name_short'])

    # if term is not easily mistaken as a name (avoid false positives)
    if term not in list_notnames:

        # if term is in the list of all names and roles
        if term in (list_all_names + list_roles):
            # get correct name and uniqueID, or role, for that term
            str_name, str_role, list_uniqueID = get_string(term, str_name, str_role, list_uniqueID, str_canton)

            if bln_print:
                print('=== correct name', term)
        # if term in list of cantons
        elif term in list_places:
            str_canton = term
        # if term is not in list_all_names
        else:
            # look for similar names based on (normalized) Damerau-Levenshtein distance
            # !!! probably need to improve this procedure
            #       - find better values ....
            if bln_print:
                print(term)
            array_normalized = array_all_names[normalized_damerau_levenshtein_distance_ndarray(term, array_all_names) <= 0.35]
            array_normalized_values = normalized_damerau_levenshtein_distance_ndarray(term, array_normalized)
            if bln_print:
                print(array_normalized, array_normalized_values)
            array_absolute = array_all_names[damerau_levenshtein_distance_ndarray(term, array_all_names) <= 2]
            array_absolute_values = damerau_levenshtein_distance_ndarray(term, array_absolute)
            if bln_print:
                print(array_absolute, array_absolute_values)
            set_intersection = set(array_normalized).intersection(set(array_absolute))
            # check if a similar name was found
            term_approx = ''
            if len(set_intersection) == 1:
                term_approx = list(set_intersection)[0]
                if bln_print:
                    print('we found the name', set_intersection)
            elif len(set_intersection) > 1:
                # !!! we only look at normalized values
                # !!! we don't account for names with same values !!!
                array_min = array_normalized[array_normalized_values.argmin()]
                term_approx = array_min#[0]
                if bln_print:
                    print('we found several possible names', set_intersection, 'and choose', array_min)
            if term_approx:
                str_name, str_role, list_uniqueID = get_string(term_approx, str_name, str_role, list_uniqueID, str_canton)
                if bln_print:
                    print('*******************', str_name, term_approx)


    return str_name, str_role, list_uniqueID, str_canton


# two functions for language identification
# Author: Luis Salamanca
# small modifications by Lili Gasser
# Using stopwords
# input:
# - text: string
# - valid_lang: tuple of valid languages
# output:
# - dict_language_counts: dictionary of stopword counts for each valid language
def identify_language(text, valid_lang = ('german', 'french', 'italian')):

    # tokenize
    tokens = text.split(' ')
    # all lowercase
    test_words = [word.lower() for word in tokens]
    # make a set
    test_words_set = set(test_words)

    # initialize dictionary of language elements
    dict_language_counts = {}

    # iterate through languages of stopwords
    for language in stopwords.fileids():
        if language in valid_lang:
            # get stopword set
            stopwords_set = set(stopwords.words(language))
            # get intersection between text of interest and stopword set for this language
            common_elements = test_words_set.intersection(stopwords_set)
            # save number of common elements to dictionary
            dict_language_counts[language] = len(common_elements)

    return dict_language_counts


# Simply, given the number of ocurrences of the stopwords, it assigns a label
# to a specific textbox, also considering the possibility of textboxes
# mixing languages. For this case, the value ratio_similar is intended
# input:
# - XML_new: XML file to update
# - aux_dict_l: corresponds to dict_language_counts
# output:
# - lang_max: string
def label_language(XML_new, ind_p, ind_t, aux_dict_l):

    # specify a similarity ratio
    ratio_similar = 0.8
    # if there are counts, determine language
    if sum(aux_dict_l.values()):
        aux_dict_l_norm = {k: v / total for total in (sum(aux_dict_l.values()),) for k, v in aux_dict_l.items()}
        lang_max_aux = max(aux_dict_l_norm.keys(), key=(lambda key: aux_dict_l_norm[key]))
        lang_max = ''
        count_l = 0
        for lang in aux_dict_l_norm.keys():
            if (aux_dict_l_norm[lang] >  aux_dict_l_norm[lang_max_aux] * ratio_similar):
                if count_l > 0:
                    lang_max += '_'
                lang_max += lang
                count_l += 1
        if count_l > 1:
            lang_max = 'mixed_' + lang_max
    else:
        lang_max = 'languageNotIdentified'

    # add attribute to textbox
    XML_new[ind_p][ind_t].attrib['language'] = lang_max

    return XML_new



















# functions from hf_extractdiscussions
# ==============================================================================
# TODO: check whether they still are needed

# function to exclude overlapping textboxes between documents
# input:
# - dict_text: dictionary of texts of one document
# - dict_overlaps_year: dictionary with overlaps
# output:
# - dict_text: modified dict_text
def exclude_overlaps(dict_text, dict_overlaps):
    # initialize to impossible values
    first_entry = -1
    last_entry = 1000

    # get index of textbox from first and last page
    # the overlap dictionary only contains an entry, if an overlap was detected
    for entry, array in dict_overlaps.items():
        if entry == 'first':
            first_entry = int(array[0])
        if entry == 'last':
            last_entry = int(array[0])

    # get list of keys for first and last page
    list_first_page = [key for key in dict_text if key.split(',')[1] == '0']
    last_page = max([int(key.split(',')[1]) for key in dict_text])
    list_last_page = [key for key in dict_text if key.split(',')[1] == str(last_page)]

    # modify dict_text on first page
    for key in list_first_page:
        if int(key.split(',')[2]) < first_entry:
            dict_text[key] = ''

    # ... and on last page
    for key in list_last_page:
        if int(key.split(',')[2]) > last_entry:
            dict_text[key] = ''

    return dict_text


# tokenizer
tokenizer_canton = RegexpTokenizer(r'\w+')    # only leaves words
#tokenizer = RegexpTokenizer(r'\w+(?:-\w+)*|\$[\d\.]+|\S+')
# last part \S+ is needed to get colon, \S stands for white space
tokenizer = RegexpTokenizer(r'\w+(?:-/w+)*|\$[\d\.]+')




# small function to get first item of tupels in a list
def get_first_item(list_tupels):
    list_first_item = [tupel[0] for tupel in list_tupels]
    return list_first_item

# small function to get last two items of tupels in a list
def get_last_item(list_tupels):
    list_last_item = [tupel[-2:] for tupel in list_tupels]
    return list_last_item






# function to get data frame from lists of names
# input:
# - lists_names: lists of names (simple, double, comp, canton)
# output:
# - df: corresponding dataframe
def get_df_from_lists_names(lists_names):
    list_types = ['simple', 'double', 'comp', 'canton']
    df = pd.DataFrame()
    for i in range(4):
        df_temp = pd.DataFrame(lists_names[i],
                columns = ('name_short', 'name_correct', 'uniqueIndex', 'CantonName', 'CantonAbbreviation', 'Citizenship', 'FirstName'))
        df_temp['type'] = list_types[i]
        df = pd.concat([df, df_temp], ignore_index = True)
    return df











def tokenize_dictionary(dictionary, tokenizer, only_text=False):
    dictionary_tokenized = {}
    # if there is only text, e.g. when we look at all texts of a document at once (level 2 in flattened dictionary)
    if only_text:
        for key, text in dictionary.items():
            dictionary_tokenized[key] = tokenizer.tokenize(text)
    # if the values are actually tuples (speaker, text), e.g. when a document corresponds to what one person said
    else:
        for key, text in dictionary.items():
            dictionary_tokenized[key] = (text[0], tokenizer.tokenize(text[1]))

    return dictionary_tokenized



# value of dictionary needs to be tokenized !!!!
def remove_stopwords_from_dictionary(dictionary, list_stopwords, only_text=False):
    dict_docs_afterswr = {}
    # if there is only text, e.g. when we look at all texts of a document at once (level 2 in flattened dictionary)
    if only_text:
        for doc, text in dictionary.items():
            list_text = text
            list_words_tokenized = [word for word in list_text if word.lower() not in list_stopwords]
            dict_docs_afterswr[doc] = ' '.join(list_words_tokenized)
    # if the values are actually tuples (speaker, text), e.g. when a document corresponds to what one person said
    else:
        for doc, tupel in dictionary.items():
            before_colon = tupel[0]
            list_text = tupel[1]
            list_words_tokenized = [word for word in list_text if word.lower() not in list_stopwords]
            dict_docs_afterswr[doc] = (before_colon, ' '.join(list_words_tokenized))

    return dict_docs_afterswr



def dict_only_text(dictionary):
    dictionary_only_text = {}
    for key, tupel in dictionary.items():
        dictionary_only_text[key] = tupel[1]

    return dictionary_only_text





# function to get list of places
def get_list_cantons(df_names):
    df_temp = df_names.loc[df_names['type']=='canton']
    list_cantonname = list(df_temp['CantonName'])
    list_cantonabbr = list(df_temp['CantonAbbreviation'])
    list_citizenship = list(df_temp['Citizenship'])
    list_citizenship = [city[:-5] for item in list_citizenship for city in item.split(',')]
    list_firstname = list(df_temp['FirstName'])

    return list_cantonname +  list_cantonabbr + list_citizenship + list_firstname
